   --------------------------------------------------------------------
--|  Program :    Jotto (word game)               Version :  1.3       
   --------------------------------------------------------------------
--|  Abstract :   Plays the solitaire word guessing game of Jotto.     
--|               Refer to the specification for a complete            
--|               description of the game.                             
   --------------------------------------------------------------------
--|  Compiler/System : Meridian Ada                                    
--|  Author          : John Dalbey                  Date :  1/93       
--|                  :                              V1.0    12/94      
--|                  : S. Ron Oliver                V1.1    Jan 94     
--|                  : John Dalbey                  V1.2    10 Feb 95  
--|                  : John Dalbey                  V1.3    Jun 95     
   --------------------------------------------------------------------
--|  Usage           : Uses a data file: jottowrd.dat                  
--|                  : Uses ANSI escape codes for showing screen colors
--|
--|  Version History : V0.1 -  Original prototype                      
--|                  : V1.0 -  Redesign Remove Duplicates routine      
--|                  : V1.1 -  Modified for conformance to style for   
--|                  :         CS1_CS2 Lab Project (D.I.S.A Funded)    
--|                  : V1.2 -  Tailored for lab exercises              
--|                  : V1.3 -  Cosmetic changes to source code layout  
   -------------------------------------------------------------------- 
WITH Text_IO;
WITH Random;
WITH Screen;
USE  Text_IO;

PROCEDURE Jotto IS

--
--   Declarations
--
TurnLimit : constant natural := 20;  -- Number of turns player allowed
DataFile  : constant STRING(1..12) := "jottowrd.dat";  -- data file name
WordSize  : constant natural := 5;   -- size of a legal word

PACKAGE My_Int_IO IS NEW Integer_IO (Integer);

------------------------------------------------------------------------------
PROCEDURE Show_Title_Screen IS
-- PURPOSE     : Display the title, instructions, column headers, and row
--             : numbers.
-- CALLED BY   : Play A Game
-- ASSUMES     : Nothing.
-- RETURNS     : Nothing.
-- ALGORITHM   :
--               Show Title Graphic
--               Show Instructions
--               Show Column Headers
--               Show Turn Numbers in a column
BEGIN

   Screen.ClearScreen;
   Screen.MoveCursor (1,1);

   -- Show Instructions for playing the game
   Screen.MoveCursor(12,45);
   Put ("Welcome to JOTTO");
   Screen.MoveCursor(14,24);
   Put ("Try to guess a secret word the computer has chosen.");
   Screen.MoveCursor(15,24);
   PUT ("After each guess the computer will give you a hint  ");
   Screen.MoveCursor(16,24);
   PUT ("which indicates how many of the letters in your guess");
   Screen.MoveCursor(17,24);
   PUT ("are also in the secret word (but not necessarily in ");
   Screen.MoveCursor(18,24);
   PUT ("the correct position).");
   Screen.MoveCursor(19,24);
   PUT ("Your guess must be 5 letters long,");
   Screen.MoveCursor(20,24);
   PUT ("     may not contain repeated letters, ");
   Screen.MoveCursor(21,24);
   PUT ("     and may not end in the letter S.");
   New_Line;
   Screen.MoveCursor(22,24);
   PUT ("You have twenty turns.  Good luck! ");

   -- Show Column Headers
   Screen.MoveCursor(1,1);
   Put_Line ("Turn   Guess   Hint");
   Put_Line ("----   -----   ----");

   -- Show turn numbers in a column
   -- (Add code here)

END Show_Title_Screen;

-------------------------------------------------------------------------------
FUNCTION Is_A_to_Z (Word: IN string) return BOOLEAN IS
-- PURPOSE     : Determine if the characters in a word are in the range A to Z.
-- CALLED BY   : Valid Word
-- ASSUMES     :
-- RETURNS     : TRUE if all characters are in range A to Z (either case).
--             : FALSE otherwise.
-- ALGORITHM   :
--       FOR each letter from Word
--          IF Letter is NOT in A to Z THEN Return FALSE

BEGIN
   RETURN TRUE;
END Is_A_to_Z;

-------------------------------------------------------------------------------
FUNCTION No_Duplicates (Word: IN String) RETURN BOOLEAN IS
-- PURPOSE     : Determine if there are any repeated letters in a word.
-- CALLED BY   : Valid Word
-- ASSUMES     : Word is a string of length 5 (that has a value).
-- RETURNS     : TRUE if there are no letter occurs twice.
--             : FALSE if a letter is duplicated (regardless of position).
-- ALGORITHM   :
--       Consider each letter from Word
--          Compare to each other letter
--          if they match (it's a duplicate) THEN return false
--

NoDuplicate: boolean;      -- Flag to indicate duplicate found

BEGIN
         NoDuplicate := TRUE;  -- start by indicating no duplicates found

         -- Consider each letter from Word
         FOR source in 1..WordSize-1 LOOP
            -- Compare to each letter in rest of word
            FOR target in source+1 .. WordSize LOOP
               -- See if they match (it's a duplicate)
               IF Word(source) = Word(target) THEN
                  NoDuplicate := FALSE;      -- show duplicate found
               END IF;
            END LOOP;
         END LOOP;

         RETURN NoDuplicate;

END No_Duplicates;

-------------------------------------------------------------------------------
FUNCTION  Valid_Word (Word: string; Length: Natural) return BOOLEAN IS
-- PURPOSE     : Determine if a word satisfies the jotto rules.
-- CALLED BY   : Get Guess, Get Secret Word
-- ASSUMES     : Word is a string.
--             : Length is the number of characters in the string.
-- RETURNS     : TRUE if
--             :    1) Length is exactly five.
--             :    2) All letters are in the range A to Z.
--             :    3) Has no repeated letters.
--             :    4) Doesn't end in the letter S.
--             : FALSE if any of the above conditions aren't met.
-- ALGORITHM   :
--               IF length = 5 AND
--                  All letters are in range A to Z AND
--                  Has no repeated letters AND
--                  Doesn't end in S
--               THEN
--                  RETURN TRUE
--               ELSE
--                  RETURN FALSE
--

BEGIN
   -- See if proper length
   IF Length /= WordSize THEN
      RETURN FALSE;
   ELSE
      --  Do all other validity checks
      RETURN Is_A_to_Z(Word) AND
             No_Duplicates(Word(1..WordSize));
   END IF;
END Valid_Word;

-------------------------------------------------------------------------------
PROCEDURE Get_Secret_Word (Secret: OUT String) IS
-- PURPOSE     : Obtain a secret word for the user to try to figure out.
-- CALLED BY   : Play A Game
-- ASSUMES     : nothing.
-- RETURNS     : Secret contains a valid word.
-- ALGORITHM   :
--                Get a word from dictionary file
--                WHILE NOT  Valid_Word   LOOP
--                      Display error message
--                      Get another word from file
--                END LOOP
--                Convert word to lower case

WordRead: String(1..WordSize);  -- input word

   -----------------------------------------------------
   PROCEDURE Get_From_File (WordfromFile: OUT string) IS
   -- PURPOSE     : Obtain a random word from the jotto dictionary file.
   -- CALLED BY   : Get Secret Word
   -- ASSUMES     : The file has one word per line, all words are 5 chars long.
   -- RETURNS     : WordFromFile contains an unvalidated word randomly selected
   --             : from the file.
   --             : If the dictionary file doesn't exist, a default word is
   --             : returned.
   -- ALGORITHM   :
   --                Open the file
   --                IF  File not found THEN use default word.
   --                REPEAT
   --                    Read a word
   --                    Increment word count
   --                UNTIL wordcount = Random Number or EOF
   --                Close the file

   theFile     : Text_IO.FILE_TYPE; -- file descriptor
   RecCount    : NATURAL;  -- how many words we've read from the file
   RandomNumber: NATURAL;  -- random word number

   BEGIN
     -- In this version, always get the first word from the file,
     -- by hard coding the computation of random number to be 1.
     RandomNumber := 1;   -- This is NOT random, but a constant,
    
     Text_IO.Open(theFile,Text_IO.In_File,DataFile);

     RecCount := 0;
     --   Read records from the file until we get to the one we want
     WHILE (NOT Text_IO.End_of_File(theFile)) AND RecCount < RandomNumber LOOP
         RecCount := RecCount + 1;
         Text_IO.Get (theFile, WordFromFile);
     END LOOP;
     Text_IO.Close(theFile);

   END Get_From_File;
   ------------------

BEGIN  -- Get_Secret_Word
   WordRead(1..WordSize) := "?????"; -- initialize word
   Get_From_File (WordRead);         -- call subprogram to obtain a word
   Secret := WordRead(1..WordSize);  -- store results for return
END Get_Secret_Word;

-------------------------------------------------------------------------------
PROCEDURE Give_Hint (Secret,
                     Guess  : IN String;
                     Turn   : IN Integer) IS
-- PURPOSE     : Display the user's guess and a hint value.
--             : The hint is the number of letters the guess has in common
--             : with the computer's word.
-- CALLED BY   : Play_A_Game
-- ASSUMES     : Secret contains the computer's secret word.
--             : Guess contains the user's validated guess.
--             : Turn contains what turn it is.
-- RETURNS     : nothing.
-- ALGORITHM   :
--               Consider each letter in secret word
--                  Consider each letter in guess word
--                     IF match then increment hint count
--               Return hint count

Hint: Integer; -- Count of common letters

BEGIN

      Hint := 0;     -- initialize Hint

      -- Consider each letter in Secret
      FOR i in 1 .. WordSize LOOP
         -- Examine each letter in Guess Word
         FOR j in 1 .. WordSize LOOP
            -- If they match increment hint count
            IF Guess(j) = Secret(i) THEN
               Hint := Hint + 1;
            END IF;
         END LOOP;
      END LOOP;

   -- Display the Guessed word and the hint
      Screen.MoveCursor(Turn+3,8);
      PUT (Guess & "     ");
      My_Int_IO.Put (Hint, WIDTH=>1);

END Give_Hint;


-------------------------------------------------------------------------------
PROCEDURE Erase_Msg IS
--
--  PURPOSE: Clear the message line at the bottom of the screen.
--  CALLED BY: Get_Guess
--
BEGIN
         Screen.MoveCursor(24,1);
         Put     ("                                        " &
                  "                                   ");
         Screen.MoveCursor (24,1);
END Erase_Msg;

-------------------------------------------------------------------------------
PROCEDURE Prompt IS
-- PURPOSE     : Display a prompt informing the user to enter a guess.
-- CALLED BY   : Get Guess
-- ASSUMES     : nothing
-- RETURNS     : nothing
-- ALGORITHM   : Clear previous input and Put the prompt to the screen.

BEGIN
         Screen.MoveCursor(23,12);
         Put("                                                      " &
             "               ");
         Screen.MoveCursor(23,7);
         PUT ("Your Guess? ");
END Prompt;


-------------------------------------------------------------------------------
PROCEDURE Get_Guess (Guess: OUT String) IS
-- PURPOSE     : Obtain a valid guess from the user.
-- CALLED BY   :
-- ASSUMES     : nothing.
-- RETURNS     : Guess is a string containing a valid jotto word.
-- ALGORITHM   :
--               REPEAT
--               Show Prompt
--               Read a line from keyboard
--               IF NOT Valid_Word THEN Show Error Message
--               UNTIL Valid_Word

WordIn: String(1..WordSize);  -- input word from user
Length: Natural := 0;         -- number of characters read by Get_Line

   -------------------------------------------------------------
   PROCEDURE Show_Error (Word: IN STRING; Length: IN NATURAL) IS
   -- PURPOSE     : Display a message describing what is invalid about the guess.
   -- CALLED BY   : Get Guess
   -- ASSUMES     : Word is a string.
   --             : Length is the number of characters in the string.
   -- RETURNS     : nothing.
   -- ALGORITHM   :
   --               Perform each validation test, and display a message
   --               on the one which fails:
   --                Length error
   --                Character range error
   --                Ends in S error
   --                Duplicate letters error
   -- NOTE: If player just pressed <Enter>, then Length = 0,
   --       so don't give an error message (we want to ignore it).

   BEGIN
   IF Length > 0 THEN
       Put ("Invalid entry: " & (Word(1..Length)) & ".  ");
   END IF;
   END Show_Error;
   ---------------

   ------------------------------------------------------
   FUNCTION IsSpecial (Word: IN STRING) RETURN BOOLEAN IS
   -- PURPOSE     : Determine if a special, undocumented word has been entered.
   -- CALLED BY   : Get Guess
   -- ASSUMES     : Word is a string.
   -- RETURNS     : TRUE if the word is "special", FALSE otherwise.
   -- ALGORITHM   :
   --               IF word is special THEN return true

   BEGIN
      RETURN  Word(1..4) = "quit" OR Word = "#####";
   END IsSpecial;
   --------------

BEGIN -- Get_Guess
   LOOP   --repeat until valid guess
      -- Display a prompt and obtain a line of input from user
      Prompt;
      Get_Line (WordIn, Length);
      Erase_Msg;

      -- If word is exactly 5 letters, gobble extra chars and <CR>
      IF Length = WordSize THEN
            Skip_Line;
      END IF;

      EXIT WHEN Valid_Word(WordIn, Length) or IsSpecial(WordIn);
      Show_Error (WordIn, Length);  -- if not valid, explain why
   END LOOP;
   Guess := WordIn;           -- Prepare to return the guess
END Get_Guess;


-------------------------------------------------------------------------------
FUNCTION Game_Over ( Secret,
                     Guess: IN String;
                     Turns: IN integer) return Boolean IS
-- PURPOSE     : Determine if the game is done, either won or lost.
-- CALLED BY   : Play A Game
-- ASSUMES     : Secret contains the computer's secret word.
--             : Guess contains the user's validated guess.
--             : Turn contains what turn it is.
-- RETURNS     : TRUE if Guess matches Secret or Turns exceeds allowed limit.
--             : FALSE otherwise
-- ALGORITHM   :
--               IF Guess = secret THEN
--                    Display Win message
--                    Return true
--               IF Out of turns THEN
--                    Display Lose message
--                    Return true
--               ELSE Return False

   ----------------------------------------
   PROCEDURE Display_Win (Turns: Integer) IS
   -- PURPOSE     : Display a congratulations message for winning the game,
   --             : which includes how many turns they took.
   -- CALLED BY   : Game Over
   -- ASSUMES     : Turns contains what turn it is.
   -- RETURNS     : nothing.
   -- ALGORITHM   :
   --               Print win message and turn count

   BEGIN
   Screen.MoveCursor(12,24);
   Put ("* * * * * * * * * * * * * * * * * * * * * * * * * * ");
   Screen.MoveCursor(13,24);
   PUT ("            Y O U   W I N !                         ");
   Screen.MoveCursor(14,24);
   Put ("* * * * * * * * * * * * * * * * * * * * * * * * * * ");
   Screen.MoveCursor(15,24);
   PUT ("You guessed the secret word in ");
   My_Int_IO.PUT (Turns,2);
   PUT (" turns.             ");
   New_Line;
   FOR i in 16 .. 22 LOOP  -- clear rest of screen
      Screen.MoveCursor(i,24);
      PUT ("                                                     ");
   END LOOP;
   New_Line;
   END Display_Win;
   ---------------

BEGIN  -- Game_Over

   Screen.MoveCursor(23,1);
   -- Win by guessing the secret word exactly
   IF Guess = Secret THEN
      Display_Win (Turns);
      RETURN TRUE;
   END IF;

   -- Lose by running out of turns
   IF Turns >= TurnLimit THEN
      PUT ("            Out of turns, you lose!  The word was: ");
      PUT (Secret);
      NEW_Line;
      RETURN TRUE;
   END IF;

   RETURN FALSE;     -- If neither win or lose, continue playing

END Game_Over;


-------------------------------------------------------------------------------
PROCEDURE Play_A_Game IS
-- PURPOSE     : Control playing one round of the Jotto.  This is the main logic
--             : for orchestrating getting guesses, giving hints, and checking
--             : if the game is over.
-- CALLED BY   : Play Jotto
-- ASSUMES     : Nothing.
-- RETURNS     : Nothing.
-- ALGORITHM   :
--                   ShowTitleScreen;
--                   Obtain a secret word
--                   REPEAT 
--                      Obtain a guess from player
--                      See if user knows the magic word
--                      Give a Hint
--                      Increment Turn count
--                   Until the game is over

Secret,                           -- The Secret Word
Guess    : STRING(1..WordSize);   -- The players guess
Turn     : INTEGER := 0;          -- Count of # of turns

BEGIN

   PUT (ASCII.ESC & "[44;33m" );   -- set screen colors
   Screen.SetAttribute(screen.bold);

   Show_Title_Screen;
   Get_Secret_Word (Secret);            -- Obtain a secret word

   -- Loop until the game is over
   WHILE NOT Game_Over (Secret, Guess, Turn) LOOP
      Guess := "     ";                 -- initialize the guess to empty
      Get_Guess (Guess);                -- Obtain a guess from player
      EXIT WHEN (Guess(1..4) = "quit"); -- User aborts
      Give_Hint (Secret, Guess, Turn);  -- Give a Hint
      Turn := Turn + 1;                 -- Increment Turn count
   END LOOP;

   PUT (ASCII.ESC & "[40;37m" );   -- set colors back to normal
   Screen.SetAttribute(screen.normal);

END Play_A_Game;



-------------------------------------------------------------------------------
BEGIN -- Jotto
   Play_A_Game;
END;
